/*
  urlwrite and urlread, a curl front-end for octave 
  Copyright (C) 2006 Alexander Barth <abarth@marine.usf.edu>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/


#include <string>
#include <fstream>
#include <iomanip>

#include <octave/oct.h>
#include "oct-env.h"
#include "ov-cell.h"

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>


// write callback function for curl

int 
write_data(void *buffer, size_t size, size_t nmemb, void *streamp)
{
  // *stream is actually an ostream object
  std::ostream& stream=*(reinterpret_cast<std::ostream*>(streamp));
  stream.write(reinterpret_cast<const char*>(buffer), size*nmemb);
  return (stream.fail() ? 0 : size * nmemb);
}

// progress callback function for curl

int 
progress_func(const char *url,double dltotal,double dlnow,double ultotal,double ulnow)
{
  // macro that picks up Ctrl-C signalling
  OCTAVE_QUIT;

# ifndef URL_QUITE_DOWNLOAD
  if (dltotal != 0) 
    {
      // is the carriage return character "\r" portable ?
      octave_stdout <<  "\r" << url << ": " << std::fixed << std::setw(5) << std::setprecision(1) << dlnow*100.0/dltotal << " %";
      octave_stdout.flush();
    }
# endif

  return 0;
}


// form the query string based on param

std::string 
form_query_string(CURL *curl,const Cell& param) {
  std::ostringstream query ;
  std::string name, text;
  char *enc_name, *enc_text;

  for (int i=0; i< param .numel(); i+=2) 
    {
      name = param(i).string_value();
      text = param(i+1).string_value();

      // encode strings
      enc_name = curl_easy_escape(curl,name.c_str(),name.length());
      enc_text = curl_easy_escape(curl,text.c_str(),text.length());

      query << enc_name << "=" << enc_text;

      curl_free(enc_name) ;
      curl_free(enc_text) ;

      if (i < param.numel()-1)
	query << "&";
    }

  query << std::flush ;
  return query.str();
}


// curl front-end

CURLcode 
urlget(const std::string& url,const std::string& method,const Cell& param,std::ostream& stream) {
  CURL *curl;

  // curl return code
  CURLcode res; 

  curl_global_init(CURL_GLOBAL_DEFAULT);

  curl = curl_easy_init();

  if (!curl) 
    return CURLE_FAILED_INIT;

  // handle paramters of GET or POST request

  std::string query_string = form_query_string(curl,param);
  //octave_stdout << "query_string " << query_string << std::endl;

  if (method == "get") 
    {
      query_string = url + "?" + query_string;
      curl_easy_setopt(curl, CURLOPT_URL,query_string.c_str());
    }
  else if (method == "post") 
    {
      curl_easy_setopt(curl, CURLOPT_URL,url.c_str());
      curl_easy_setopt(curl, CURLOPT_POSTFIELDS, query_string.c_str());
    }
  else 
    {
      curl_easy_setopt(curl, CURLOPT_URL,url.c_str());
    }

  // Define our callback to get called when there's data to be written
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_data);
  // Set a pointer to our struct to pass to the callback
  curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*)&stream);

  curl_easy_setopt(curl, CURLOPT_NOPROGRESS, false);
  curl_easy_setopt(curl, CURLOPT_PROGRESSFUNCTION, progress_func);
  curl_easy_setopt(curl, CURLOPT_PROGRESSDATA, url.c_str());
  curl_easy_setopt(curl, CURLOPT_FAILONERROR, true);
  // Switch on full protocol/debug output
  // curl_easy_setopt(curl, CURLOPT_VERBOSE, true);

  res = curl_easy_perform(curl);

# ifndef URL_QUITE_DOWNLOAD
  if (res == CURLE_OK) 
    {
      // download is complete
      progress_func(url.c_str(),1,1,0,0);
      // new line after progress_func
      octave_stdout << std::endl;
    }
# endif

  // always cleanup
  curl_easy_cleanup(curl);

  curl_global_cleanup();

  return res;
}


DEFUN_DLD (urlwrite, args, nargout,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {} urlwrite (@var{URL},@var{localfile})\n\
@deftypefnx {Loadable Function} {@var{f} =} urlwrite(@var{url},@var{localfile})\n\
@deftypefnx {Loadable Function} {[@var{f},@var{success}] =} urlwrite (@var{url},@var{localfile})\n\
@deftypefnx {Loadable Function} {[@var{f},@var{success},@var{message}] =} urlwrite(@var{url},@var{localfile})\n\
Downloads a remote file specified by its @var{URL} and saves it as @var{localfile}. \n\
Example: \n\n\
@example \n\
urlwrite('ftp://ftp.octave.org/pub/octave/README','README.txt'); \n\
@end example \n\n\
The full path of the downloaded file is returned in @var{f}. The variable \n\
@var{success} is 1 if the download was successful, otherwise it \n\
is 0 in which case @var{message} contains an error message. If no output argument \n\
is specified and if an error occurs, then the error is signaled through \n\
octave's error handling mechanism. \n\
\n\
This function uses libcurl. Curl supports, among others, the HTTP, FTP and FILE protocols. \n\
Username and password may be specified in the URL, for example: \n\n\
@example \n\
urlwrite('http://username:password@@example.com/file.txt','file.txt'); \n\
@end example \n\n\
GET and POST requests can be specified by @var{method} and @var{param}. @var{method} is either \n\
'get' or 'post' and @var{param} is a \n\
cell array of parameter and value pairs, for example: \n\n\
@example \n\
urlwrite('http://www.google.com/search','search.html','get',@{'query','octave'@});\n\
@end example\n\
@end deftypefn\n\
@seealso{urlread}")
{
  // curl return code
  CURLcode res; 

  // octave's return value
  octave_value_list retval;

  int nargin = args.length ();

  // verify arguments
  if (nargin != 2 && nargin != 4) 
    {
      print_usage ();
      return retval;
    }

  if (!args(0).is_string()) 
    {
      error("urlwrite: url must be a string argument");
      return retval;
    }

  std::string url = args(0).string_value();

  if (!args(1).is_string()) 
    {
      error("urlwrite: localfile must be a string argument");
      return retval;
    }

  // name to store the file if download is succesful 
  std::string filename = args(1).string_value();

  std::string method;
  Cell param; // empty cell array

  if (nargin == 4) 
    {
      if (!args(2).is_string()) 
        {
          error("urlwrite: method can only be \"get\" or \"post\".");
          return retval;
        }

      method = args(2).string_value();

      if (method != "get" && method != "post") 
	{
	  error("urlwrite: method can only be \"get\" or \"post\".");
	  return retval;
	}

      if (!args(3).is_cell()) 
	{
	  error("urlwrite: parameters for get and post requests must be given as a cell.");
	  return retval;
	}

      param = args(3).cell_value();

      if (param.numel() % 2 == 1 ) 
	{
	  error("urlwrite: number of elements in param must be even.");
	  return retval;
	}

    }

  // create and open file stream

  std::ofstream stream(filename.c_str(), std::ios::out | std::ios::binary);

  if (!stream.is_open()) 
    {
      error("urlwrite: unable to open file.");
      return retval;
    }

  res = urlget(url,method,param,stream);

  // close the local file
  stream.close();


  if(nargout > 0) 
    {
      // FIXME: urlwrite should return full file path
      retval(0) = octave_env::make_absolute(filename, octave_env::getcwd ());
      //      retval(0) = filename;
      retval(1) = res == CURLE_OK;
      // return empty string if no error occured
      retval(2) = std::string(res == CURLE_OK ? "" : curl_easy_strerror(res));
    }

  if(nargout < 2 & res != CURLE_OK) 
    {
      error("curl: %s\n", curl_easy_strerror(res));
    }

  return retval;
}



DEFUN_DLD (urlread, args, nargout,
	   "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {@var{s} =} urlread(@var{url})\n\
@deftypefnx {Loadable Function} {[@var{s},@var{success}] =} urlread (@var{url})\n\
@deftypefnx {Loadable Function} {[@var{s},@var{success},@var{message}] =} urlread(@var{url})\n\
@deftypefnx {Loadable Function} {[...] =} urlread(@var{url},@var{method},@var{param})\n\
Downloads a remote file specified by its @var{URL} and return its content in string @var{s}. Example: \n\n\
@example \n\
s = urlread('ftp://ftp.octave.org/pub/octave/README'); \n\
@end example \n\n\
The variable @var{success} is 1 if the download was successful, otherwise it \n\
is 0 in which case @var{message} contains an error message. If no output argument \n\
is specified and if an error occurs, then the error is signaled through \n\
octave's error handling mechanism. \n\
\n\
This function uses libcurl. Curl supports, among others, the HTTP, FTP and FILE protocols. \n\
Username and password may be specified in the URL, for example: \n\n\
@example \n\
s = urlread('http://username:password@@example.com/file.txt'); \n\
@end example \n\n\
GET and POST requests can be specified by @var{method} and @var{param}. @var{method} is either \n\
'get' or 'post' and @var{param} is a \n\
cell array of parameter and value pairs, for example: \n\n\
@example \n\
s = urlread('http://www.google.com/search','get',@{'query','octave'@});\n\
@end example\n\
@end deftypefn\n\
@seealso{urlwrite}")
{
  // curl return code
  CURLcode res; 

  // octave's return value
  octave_value_list retval;

  int nargin = args.length ();

  // verify arguments
  if (nargin != 1  && nargin != 3) 
    {
      print_usage ();
      return retval;
    }

  if (!args(0).is_string()) 
    {
      error("urlread: url must be a string argument");
      return retval;
    }

  std::string url = args(0).string_value();

  std::string method;
  Cell param; // empty cell array

  if (nargin == 3) 
    {
      if (!args(1).is_string()) 
	{
	  error("urlread: method can only be \"get\" or \"post\".");
	  return retval;
	}

      method = args(1).string_value();

      if (method != "get" && method != "post") 
	{
	  error("urlread: method can only be \"get\" or \"post\".");
	  return retval;
	}

      if (!args(2).is_cell()) 
	{
	  error("urlread: parameters for get and post requests must be given as a cell.");
	  return retval;
	}

      param = args(2).cell_value();

      if (param.numel() % 2 == 1 ) 
	{
	  error("urlread: number of elements in param must be even.");
	  return retval;
	}

    }

  // string stream for output
  std::ostringstream stream;

  res = urlget(url,method,param,stream);

  if(nargout > 0) 
    {
      retval(0) = stream.str();
      retval(1) = res == CURLE_OK;
      // return empty string if no error occured
      retval(2) = std::string(res == CURLE_OK ? "" : curl_easy_strerror(res));
    }

  if(nargout < 2 & res != CURLE_OK) 
    {
      error("curl: %s\n", curl_easy_strerror(res));
    }

  return retval;
}
